from math import isqrt
from functools import partial 
from .integer_pairing._interface import abstractPair, abstractUnpair

def _pair(m, n):
    return n*n+m if n > m else m*m+m*n

def _unpair(n):
   q = isqrt(n)
   l = n - q*q 
   return (l, q) if l < q else (q, l-q) 

pair = partial(abstractPair, pair_fun=_pair)
unpair = partial(abstractUnpair, unpair_fun=_unpair)


from math import isqrt
from functools import partial
from .integer_pairing._interface import abstractPair, abstractUnpair

def _pair(m, n):
    return (m + n) * (m + n + 1) // 2 + n

def _unpair(z):
    a = 8 * z + 1
    m = isqrt(100 * a // 4)
    w = m // 10
    if m % 10 < 5: w -= 1
    t = (w * w + w) // 2
    y = z - t
    x = w - y
    return x, y

pair = partial(abstractPair, pair_fun=_pair)
unpair = partial(abstractUnpair, unpair_fun=_unpair)



from functools import reduce

def _encodeNegatives(l):
    return [2*x if x>=0 else -2*x+1 for x in l]

def _decodeNegatives(l):
    return [x//2 if x%2==0 else -(x-1)//2 for x in l]

def abstractPair(*args, pair_fun):
    assert args, "you have to pair something"
    assert all(isinstance(x, int) for x in args)

    for x in args:
        if x < 0:
            args = _encodeNegatives(args)
            break

    if len(args) == 1: return args[0]
    return reduce(pair_fun, args)

def abstractUnpair(n, dim=2, neg=False, unpair_fun=False):
    assert isinstance(n, int)
    assert isinstance(dim, int)
    assert dim >= 1
    
    p = []
    for _ in range(dim-1):
        n, m = unpair_fun(n)
        p.append(m)
    p.append(n)
    
    p.reverse()
    if neg: p = _decodeNegatives(p)
    return tuple(p)